'use strict'
// @ts-check

const _ = require('lodash')
const config = require('config')
const amqplib = require('amqplib')
const { isFatalError } = require('amqplib/lib/connection')
const Logger = require('@fizz.js/node-logger')
const Common = require('@fizz.js/node-common')

/**
 *
 *
 * @class Rabbitmq
 */
class Rabbitmq {
  /**
   * Creates an instance of Rabbitmq.
   *
   * @memberof Rabbitmq
   */
  constructor() {
    /** @type {IRabbitmqOptions} */
    const options = config.rabbitmq
    if (_.isEmpty(options)) {
      throw new TypeError('invalid config for rabbitmq!')
    }

    this.options = options
    this.mq = amqplib
    this.connection = undefined
  }

  /**
   *
   *
   * @param {Error} err
   * @returns {boolean}
   * @memberof Rabbitmq
   */
  isFatalError(err) {
    return isFatalError(err)
  }

  /**
   *
   *
   * @param {Error} err
   * @returns {boolean}
   * @memberof Rabbitmq
   */
  isIllegalOperationError(err) {
    return _.get(err, 'name') === 'IllegalOperationError'
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  initConnectionListener() {
    try {
      if (this.connectionListeners) {
        this.removeAllConnectionListeners()
      }

      this.connectionListeners = new Map()
      this.addConnectionListener('error', this.onConnectionError.bind(this))
      this.addConnectionListener('close', this.onConnectionClose.bind(this))
      this.addConnectionListener('blocked', this.onBlocked.bind(this))
      this.addConnectionListener('unblocked', this.onUnblocked.bind(this))
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @param {string} message
   * @param {Function} callback
   * @returns {void}
   * @memberof Rabbitmq
   */
  addConnectionListener(message, callback) {
    try {
      if (this.connectionListeners) {
        if (this.connectionListeners.has(message)) {
          this.removeConnectionListener(message, callback)
        }

        this.connectionListeners.set(message, callback)
      }

      if (this.connection) {
        this.connection.addListener(message, callback)
      }
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @param {string} message
   * @param {Function} callback
   * @returns {void}
   * @memberof Rabbitmq
   */
  removeConnectionListener(message, callback) {
    try {
      if (this.connectionListeners) {
        this.connectionListeners.delete(message)
      }

      if (this.connection) {
        this.connection.removeListener(message, callback)
        this.connection.removeAllListeners(message)
      }
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  removeAllConnectionListeners() {
    try {
      if (this.connectionListeners) {
        ;[...this.connectionListeners.keys()].map(message =>
          this.removeConnectionListener(message, this.connectionListeners.get(message))
        )
        this.connectionListeners.clear()
      }
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  initChannelListener() {
    try {
      if (this.channelListeners) {
        this.removeAllChannelListeners()
      }

      this.channelListeners = new Map()
      this.addChannelListener('error', this.onChannelError.bind(this))
      this.addChannelListener('close', this.onChannelClose.bind(this))
      this.addChannelListener('return', this.onChannelReturned.bind(this))
      this.addChannelListener('drain', this.onChannelDrain.bind(this))
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @param {string} message
   * @param {Function} [callback]
   * @returns {void}
   * @memberof Rabbitmq
   */
  addChannelListener(message, callback) {
    try {
      if (this.channelListeners) {
        if (this.channelListeners.has(message)) {
          this.removeChannelListener(message, callback)
        }

        this.channelListeners.set(message, callback)
      }

      if (this.channel) {
        this.channel.addListener(message, callback)
      }
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @param {string} message
   * @param {Function} callback
   * @returns {void}
   * @memberof Rabbitmq
   */
  removeChannelListener(message, callback) {
    try {
      if (this.channelListeners) {
        this.channelListeners.delete(message)
      }

      if (this.channel) {
        this.channel.removeListener(message, callback)
        this.channel.removeAllListeners(message)
      }
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  removeAllChannelListeners() {
    try {
      if (this.channelListeners) {
        ;[...this.channelListeners.keys()].map(message =>
          this.removeChannelListener(message, this.channelListeners.get(message))
        )
        this.channelListeners.clear()
      }
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @param {*} reason
   * @returns {void}
   * @memberof Rabbitmq
   */
  onChannelDrain() {
    Logger.info('Rabbitmq.onChannelDrain')
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  onChannelReturned(msg) {
    Logger.info(`Rabbitmq.onChannelReturned: ${msg}`)
  }

  /**
   *
   *
   * @param {Error} err
   * @returns {void}
   * @memberof Rabbitmq
   */
  onChannelError(err) {
    Logger.error(err)
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  async onChannelClose() {
    try {
      Logger.warn('Rabbitmq channel is now closed...')
      await this.closeChannel()
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @param {*} reason
   * @returns {void}
   * @memberof Rabbitmq
   */
  onBlocked(reason) {
    Logger.warn(`Rabbitmq.onBlocked reason: ${reason}`)
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  onUnblocked() {
    Logger.info('Rabbitmq.onUnblocked')
  }

  /**
   *
   *
   * @param {Error} err
   * @returns {void}
   * @memberof Rabbitmq
   */
  onConnectionError(err) {
    Logger.warn(err)
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  async onConnectionClose() {
    try {
      Logger.warn('Rabbitmq connection is now closed...')
      await this.disconnect()
    } catch (error) {
      Logger.error(error)
    }
  }

  /**
   *
   *
   * @returns {amqplib.Connection}
   * @memberof Rabbitmq
   */
  async connect() {
    try {
      if (this.connection) {
        return this.connection
      }

      const url = _.get(this.options, 'url')
      if (_.isEmpty(url)) {
        throw new TypeError('invalid connection url!', this.options)
      }

      this.connection = await this.mq.connect(url)
      this.initConnectionListener()
      await this.createChannel()

      return this.connection
    } catch (error) {
      Logger.error(error)
      this.connection = undefined
      return this.connection
    }
  }

  /**
   *
   * @see https://github.com/squaremo/amqp.node/blob/v0.4.2/lib/connection.js#L364
   * @returns {boolean}
   * @memberof Rabbitmq
   */
  async disconnect() {
    try {
      if (this.connection) {
        await this.connection.close()
      }

      return true
    } catch (error) {
      if (!this.isIllegalOperationError(error) && this.isFatalError(error)) {
        Logger.error(error)
        throw error
      }

      return false
    } finally {
      this.removeAllConnectionListeners()
      this.connection = undefined
    }
  }

  /**
   *
   *
   * @returns {boolean}
   * @memberof Rabbitmq
   */
  async dispose() {
    try {
      this.removeAllChannelListeners()
      this.removeAllConnectionListeners()
      await this.closeChannel().catch(e => Logger.warn(e))
      await this.disconnect().catch(e => Logger.warn(e))
      return true
    } catch (error) {
      Logger.error(error)
      return false
    } finally {
      this.channel = undefined
      this.connection = undefined
    }
  }

  /**
   *
   *
   * @returns {Promise<amqplib.Channel>}
   * @memberof Rabbitmq
   */
  async createChannel() {
    try {
      this.channel = await this.connection.createChannel()
      this.initChannelListener()
      return this.channel
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   *
   * @returns {Promise<amqplib.ConfirmChannel>}
   * @memberof Rabbitmq
   */
  async createConfirmChannel() {
    try {
      return await this.connection.createConfirmChannel()
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  async closeChannel() {
    try {
      if (this.channel) {
        await this.channel.close()
      }
    } catch (error) {
      if (!this.isIllegalOperationError(error) && this.isFatalError(error)) {
        Logger.error(error)
        throw error
      }
    } finally {
      this.removeAllChannelListeners()
      this.channel = undefined
    }
  }

  /**
   *
   * @param {string} queue
   * @param {Object} [options={ durable: true }]
   * @returns {amqplib.Replies.AssertQueue}
   * @memberof Rabbitmq
   */
  async assertQueue(
    queue,
    options = {
      durable: true
    }
  ) {
    try {
      await this.connect()
      return await this.channel.assertQueue(queue, options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {Object} [options={ exclusive: true }]
   * @returns {amqplib.Replies.AssertQueue}
   * @memberof Rabbitmq
   */
  async assertTemporaryQueue(
    options = {
      exclusive: true
    }
  ) {
    try {
      return await this.assertQueue('', options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {*} content `content` should be Buffer.
   * @param {Object} [options={ presistent: true }]
   * @returns {boolean}
   * @memberof Rabbitmq
   */
  async sendToQueue(
    queue,
    content,
    options = {
      presistent: true,
      replyTo: undefined
    }
  ) {
    try {
      await this.connect()
      return await this.channel.sendToQueue(queue, this.convertToBuffer(content), options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} exchange
   * @param {'direct'|'fanout'|'topic'|'headers'|'default'} type
   * @param {amqplib.Options.AssertExchange} options
   * @returns {amqplib.Replies.AssertExchange}
   * @memberof Rabbitmq
   */
  async assertExchange(
    exchange,
    type = 'direct',
    options = {
      durable: true
    }
  ) {
    try {
      await this.connect()
      return await this.channel.assertExchange(exchange, type, options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} exchange
   * @param {amqplib.Options.AssertExchange} options
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async checkExchange(exchange) {
    try {
      await this.connect()
      return await this.channel.checkExchange(exchange)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} exchange
   * @param {amqplib.Options.DeleteExchange} [options = { ifUnused: true }]
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async deleteExchange(
    exchange,
    options = {
      ifUnused: true
    }
  ) {
    try {
      await this.connect()
      return await this.channel.deleteExchange(exchange, options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} destination
   * @param {string} exchange
   * @param {string} routingKey
   * @param {*} args
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async bindExchange(destination, exchange, routingKey, args = undefined) {
    try {
      await this.connect()
      return await this.channel.bindExchange(destination, exchange, routingKey, args)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} destination
   * @param {string} exchange
   * @param {string} routingKey
   * @param {*} args
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async unbindExchange(destination, exchange, routingKey, args = undefined) {
    try {
      await this.connect()
      return await this.channel.unbindExchange(destination, exchange, routingKey, args)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} exchange
   * @param {string} routingKey
   * @param {*} content `content` should be Buffer.
   * @param {amqplib.Options.Publish} options
   * @returns {boolean}
   * @memberof Rabbitmq
   */
  async publish(
    exchange,
    routingKey,
    content,
    options = {
      persistent: true,
      priority: 1
    }
  ) {
    try {
      await this.connect()
      return await this.channel.publish(exchange, routingKey, this.convertToBuffer(content), options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {string} exchange
   * @param {string} routingKey
   * @param {*} [args]
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async bindQueue(queue, exchange, routingKey, args = undefined) {
    try {
      await this.connect()
      return await this.channel.bindQueue(queue, exchange, routingKey, args)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {string} exchange
   * @param {string} routingKey
   * @param {*} [args = undefined]
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async unbindQueue(queue, exchange, routingKey, args = undefined) {
    try {
      await this.connect()
      return await this.channel.unbindQueue(queue, exchange, routingKey, args)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @returns {amqplib.Replies.AssertQueue}
   * @memberof Rabbitmq
   */
  async checkQueue(queue) {
    try {
      await this.connect()
      return await this.channel.checkQueue(queue)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {amqplib.Options.DeleteQueue} [options = { messageCount: 1 }]
   * @returns {amqplib.Replies.DeleteQueue}
   * @memberof Rabbitmq
   */
  async deleteQueue(
    queue,
    options = {
      messageCount: 1
    }
  ) {
    try {
      await this.connect()
      return await this.channel.deleteQueue(queue, options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @returns {amqplib.Replies.PurgeQueue}
   * @memberof Rabbitmq
   */
  async purgeQueue(queue) {
    try {
      await this.connect()
      return await this.channel.purgeQueue(queue)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {Function} [onMessage = args => {}]
   * @param {amqplib.Options.Consume} [options={ noAck: false }]
   * @returns {amqplib.Replies.Consume}
   * @memberof Rabbitmq
   */
  async consume(
    queue,
    onMessage = args => {},
    options = {
      noAck: false
    }
  ) {
    try {
      await this.connect()
      return await this.channel.consume(queue, onMessage, options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {amqplib.Options.Get} [options={ noAck: false }]
   * @returns {amqplib.GetMessage|false}
   * @memberof Rabbitmq
   */
  async get(
    queue,
    options = {
      noAck: false
    }
  ) {
    try {
      await this.connect()
      return await this.channel.get(queue, options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} queue
   * @param {Function} [onMessage = args => {}]
   * @param {amqplib.Options.Consume} [options={ noAck: false }]
   * @returns {*}
   * @memberof Rabbitmq
   */
  async subscribe(
    queue,
    onMessage = args => {},
    options = {
      noAck: false
    }
  ) {
    try {
      return await this.consume(queue, onMessage, options)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   * @description `ACK` というのは、応答。例えば、`cosume`による購読をして、そのジョブに対して`ack`で応答することで`queue`を消費したとみなす。 `{ noAck: true }` なら、応答の必要がない。
   * @param {amqplib.Message} message
   * @param {boolean} allUpTo
   * @returns {void}
   * @memberof Rabbitmq
   */
  async ack(message, allUpTo = false) {
    try {
      await this.connect()
      return this.channel.ack(message, allUpTo)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @returns {void}
   * @memberof Rabbitmq
   */
  async ackAll() {
    try {
      await this.connect()
      return this.channel.ackAll()
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {amqplib.Message} message
   * @param {boolean} allUpTo
   * @param {boolean} requeue
   * @returns {void}
   * @memberof Rabbitmq
   */
  async nack(message, allUpTo = false, requeue = true) {
    try {
      await this.connect()
      return this.channel.nack(message, allUpTo, requeue)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {boolean} requeue
   * @returns {void}
   * @memberof Rabbitmq
   */
  async nackAll(requeue = false) {
    try {
      await this.connect()
      return this.channel.nackAll(requeue)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @description チャンネルの受け取るメッセージ数の上限を指定する
   * @param {number} count
   * @param {boolean} isGlobal
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async prefetch(count = 1, isGlobal = false) {
    try {
      await this.connect()
      return await this.channel.prefetch(count, isGlobal)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {amqplib.Message} message
   * @param {boolean} requeue
   * @returns {void}
   * @memberof Rabbitmq
   */
  async reject(message, requeue = true) {
    try {
      await this.connect()
      return await this.channel.reject(message, requeue)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async recover() {
    try {
      await this.connect()
      return await this.channel.recover()
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   * @param {string} consumerTag
   * @returns {amqplib.Replies.Empty}
   * @memberof Rabbitmq
   */
  async cancel(consumerTag) {
    try {
      await this.connect()
      return await this.channel.cancel(consumerTag)
    } catch (error) {
      Logger.error(error)
      throw error
    }
  }

  /**
   *
   *
   * @param {*} content
   * @returns {Buffer}
   * @memberof Rabbitmq
   */
  convertToBuffer(content) {
    try {
      if (_.isObjectLike(content)) {
        content = Common.toJSON(content) || content
      }

      if (!_.isBuffer(content)) {
        content = Buffer.from(content)
      }

      return content
    } catch (error) {
      Logger.error(error)
      return content
    }
  }

  /**
   *
   *
   * @param {amqplib.Message} message
   * @returns {Object}
   * @memberof Rabbitmq
   */
  parseContent(message) {
    try {
      if (_.has(message, 'content')) {
        const content = _.get(message, 'content')
        if (_.isBuffer(content)) {
          return Common.fromJSON(_.toString(content)) || message
        }
      }

      return message
    } catch (error) {
      Logger.error(error)
      return message
    }
  }
}

module.exports = new Rabbitmq()
